/* global DOMParser, XMLSerializer */

const domElement$1 = str => {
  const parser = new DOMParser();
  const doc = parser.parseFromString(str.trim(), 'text/html');
  return doc.body.firstChild
};

function querySelectorFromString (htmlString, cssSelector) {
  if (!cssSelector) {
    return htmlString
  }
  const serializer = new XMLSerializer();
  const parser = new DOMParser();
  const doc = parser.parseFromString(htmlString, 'text/html');
  const snippet = doc.querySelector(cssSelector);
  return serializer.serializeToString(snippet)
}

const debounce = function (fn, d) {
  let timer;
  return function () {
    const context = this; const args = arguments;
    clearTimeout(timer);
    timer = setTimeout(() => {
      fn.apply(context, args);
    }, d);
  }
};

/* global HTMLElement, customElements, FormData, fetch, window , CustomEvent */


const inputElement = ({ placeholder } = {}) => domElement$1(`
  <input type="text" tabindex="1" name="combined-term" min-length="3" ${placeholder ? `placeholder="${placeholder}"` : ''} />
`);

const closeButton$1 = () => domElement$1('<button class="btn btn-sm close-button" tabindex="2">X</button>');

const submitButton = () => domElement$1(`
  <button class="btn">&gt;&gt;</button>
`);

class CombinedSearchForm extends HTMLElement {
  constructor () {
    super();
    this.input = inputElement({ placeholder: this.hasAttribute('label') ? this.getAttribute('label') : '' });
    this.submit = submitButton();
    this.closeButton = closeButton$1();
  }

  connectedCallback () {
    this.forms = this.querySelectorAll('async-search-form');
    this.formInputs = this.querySelectorAll('async-search-form input[type=text]');
    this.resultView = this.querySelector('combined-search-result-view');
    this.resultView.prepend(this.closeButton);

    if (this.hasSearchHint()) {
      this.resultView.prepend(domElement$1(`<h5 class="">${this.searchHint}</h5>`));
    }

    this.prepend(this.input);

    this.addEventListener('async-form-data-received', this.activate.bind(this));

    const debouncedSearch = debounce(this.triggerSearch.bind(this), 300);
    this.input.addEventListener('input', debouncedSearch);

    // this.input.addEventListener('focusout', this.deactivate.bind(this))
    this.input.addEventListener('focus', this.activate.bind(this));
    this.closeButton.addEventListener('click', this.deactivate.bind(this));
  }

  triggerSearch () {
    this.formInputs.forEach(input => {
      input.value = this.input.value;
    });

    // TODO: use min-length attribute on combined-search-form instead of constant 3
    if (this.input.value && this.input.value.length >= 3) {
      this.forms.forEach(form => form.submit());
    } else {
      // this.deactivate()
      this.forms.forEach(form => form.cleanup());
    }
  }

  activate () {
    this.resultView.classList.add('active');
  }

  deactivate () {
    setTimeout(() =>
      this.resultView.classList.remove('active'), 100);
  }

  get searchHint () {
    return this.getAttribute('search-hint')
  }

  hasSearchHint () {
    return this.hasAttribute('search-hint')
  }
}

const isFQDN = (urlOrPath = '') => urlOrPath.indexOf('://') > 0 || urlOrPath.indexOf('//') === 0;

class AsyncSearchForm extends HTMLElement {
  constructor () {
    super();

    this.form = this.querySelector('form');
    this.action = this.form.getAttribute('action');
    this.method = this.form.getAttribute('method');
    this.select = this.getAttribute('select');
  }

  connectedCallback () {
    this.target = this.querySelector('async-search-form-results');

    if (this.method !== 'GET') {
      console.error('async form is only supported for forms using the GET method');
    } else {
      this.form.addEventListener('submit', async ev => {
        ev.preventDefault();
        let searchUrl;
        try {
          searchUrl = isFQDN(this.action) ? this.action : this.searchUrl(ev.target);
          const response = await fetch(searchUrl);

          if (!response.ok) {
            if (response.status === 401) {
              console.error('user is logged out, refresh main resource, resulting in a redirect to the Login Provider');
              window.location.reload();
              return
            }

            const message = `could not retrieve data from ${this.action}`;
            console.error(message);
            this.renderError(message);
            return
          }

          const results = await response.text();

          const dataReceivedEvent = new CustomEvent('async-form-data-received', { bubbles: true });
          this.dispatchEvent(dataReceivedEvent);

          this.render(results);
        } catch (e) {
          const message = `failed to load resource "${searchUrl.toString()}": ${e.message}`;
          console.log(message, e);
          this.renderError(message);
        }
      });
    }
  }

  render (htmlString) {
    if (this.target) {
      this.target.innerHTML = querySelectorFromString(htmlString, this.select);
    } else {
      console.log('result from async form', querySelectorFromString(htmlString, this.select));
    }
  }

  renderError (error) {
    if (this.target) {
      this.target.innerHTML = `<p class="alert alert-danger">${error}</p>`;
    }
  }

  searchUrl (form) {
    const queryParams = new FormData(form);
    const query = new URLSearchParams(queryParams).toString();
    const searchUrl = new URL(window.location.href);
    searchUrl.pathname = this.action;
    searchUrl.search = query;
    return searchUrl
  }

  submit () {
    this.form.requestSubmit();
  }

  cleanup () {
    this.target.innerHTML = '';
  }
}

customElements.define('async-search-form', AsyncSearchForm);
customElements.define('combined-search-form', CombinedSearchForm);

/* global customElements, HTMLElement */

class ReactiveFilterSearch extends HTMLElement {
  connectedCallback () {
    this.form = this.querySelector('form');
    this.submit = this.querySelector('form input[type=submit]');
    this.submit.classList.add('hidden');

    this.changeHandler = e => {
      const { target } = e;
      const inputFieldId = target.getAttribute('id');
      if (inputFieldId) {
        // add input id hash to action, to maintain scroll height after filtering
        this.form.action = urlWithHash(this.form.action, inputFieldId);
      }
      this.form.submit();
    };

    this.addEventListener('input', this.changeHandler);
  }

  disconnectedCallback () {
    this.removeEventListener('input', this.changeHandler);
  }
}

customElements.define('reactive-filter-search', ReactiveFilterSearch);

function urlWithHash (url, hash) {
  let relativeURL, absoluteURL;
  try {
    absoluteURL = new URL(url);
  } catch (e) {
    relativeURL = url;
  }
  const a = document.createElement('a');
  a.href = absoluteURL || relativeURL;
  a.hash = hash;
  return absoluteURL ? a.href : `${a.pathname}${a.hash}`
}

/* global HTMLButtonElement, customElements */

class DateSelect extends HTMLButtonElement {
  connectedCallback () {
    this.date = this.getAttribute('date');
    this.dateField = this.closest('.form-group').querySelector('input[type=date]');

    if (this.date && this.dateField) {
      this.addEventListener('click', (e) => {
        e.preventDefault();
        this.dateField.value = this.date;
      });
    }
  }
}

customElements.define('date-select', DateSelect, { extends: 'button' });

/* global customElements, HTMLElement, DOMParser */

const domElement = str => {
  const parser = new DOMParser();
  const doc = parser.parseFromString(str, 'text/html');
  return doc.body.firstChild
};

class RangeInput extends HTMLElement {
  connectedCallback () {
    const unit = this.getAttribute('unit');
    const input = this.querySelector('input[type="range"]');
    if (input) {
      const id = input.getAttribute('id');
      const name = input.getAttribute('name');

      const output = domElement(`<output for="${id || name}">`);
      output.value = input.value;

      this.addEventListener('change', (e) => {
        output.value = input.value;
      });

      if (unit) {
        const unitWrapper = domElement(`<span>${unit}</span>`);
        unitWrapper.prepend(output);
        input.after(unitWrapper);
      } else {
        input.after(output);
      }
    }
  }
}

customElements.define('range-input', RangeInput);

/**
 * Fuse.js v6.5.3 - Lightweight fuzzy-search (http://fusejs.io)
 *
 * Copyright (c) 2021 Kiro Risk (http://kiro.me)
 * All Rights Reserved. Apache Software License 2.0
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 */

function isArray(value) {
  return !Array.isArray
    ? getTag(value) === '[object Array]'
    : Array.isArray(value)
}

// Adapted from: https://github.com/lodash/lodash/blob/master/.internal/baseToString.js
const INFINITY = 1 / 0;
function baseToString(value) {
  // Exit early for strings to avoid a performance hit in some environments.
  if (typeof value == 'string') {
    return value
  }
  let result = value + '';
  return result == '0' && 1 / value == -INFINITY ? '-0' : result
}

function toString(value) {
  return value == null ? '' : baseToString(value)
}

function isString(value) {
  return typeof value === 'string'
}

function isNumber(value) {
  return typeof value === 'number'
}

// Adapted from: https://github.com/lodash/lodash/blob/master/isBoolean.js
function isBoolean(value) {
  return (
    value === true ||
    value === false ||
    (isObjectLike(value) && getTag(value) == '[object Boolean]')
  )
}

function isObject(value) {
  return typeof value === 'object'
}

// Checks if `value` is object-like.
function isObjectLike(value) {
  return isObject(value) && value !== null
}

function isDefined(value) {
  return value !== undefined && value !== null
}

function isBlank(value) {
  return !value.trim().length
}

// Gets the `toStringTag` of `value`.
// Adapted from: https://github.com/lodash/lodash/blob/master/.internal/getTag.js
function getTag(value) {
  return value == null
    ? value === undefined
      ? '[object Undefined]'
      : '[object Null]'
    : Object.prototype.toString.call(value)
}

const EXTENDED_SEARCH_UNAVAILABLE = 'Extended search is not available';

const LOGICAL_SEARCH_UNAVAILABLE = 'Logical search is not available';

const INCORRECT_INDEX_TYPE = "Incorrect 'index' type";

const LOGICAL_SEARCH_INVALID_QUERY_FOR_KEY = (key) =>
  `Invalid value for key ${key}`;

const PATTERN_LENGTH_TOO_LARGE = (max) =>
  `Pattern length exceeds max of ${max}.`;

const MISSING_KEY_PROPERTY = (name) => `Missing ${name} property in key`;

const INVALID_KEY_WEIGHT_VALUE = (key) =>
  `Property 'weight' in key '${key}' must be a positive integer`;

const hasOwn = Object.prototype.hasOwnProperty;

class KeyStore {
  constructor(keys) {
    this._keys = [];
    this._keyMap = {};

    let totalWeight = 0;

    keys.forEach((key) => {
      let obj = createKey(key);

      totalWeight += obj.weight;

      this._keys.push(obj);
      this._keyMap[obj.id] = obj;

      totalWeight += obj.weight;
    });

    // Normalize weights so that their sum is equal to 1
    this._keys.forEach((key) => {
      key.weight /= totalWeight;
    });
  }
  get(keyId) {
    return this._keyMap[keyId]
  }
  keys() {
    return this._keys
  }
  toJSON() {
    return JSON.stringify(this._keys)
  }
}

function createKey(key) {
  let path = null;
  let id = null;
  let src = null;
  let weight = 1;

  if (isString(key) || isArray(key)) {
    src = key;
    path = createKeyPath(key);
    id = createKeyId(key);
  } else {
    if (!hasOwn.call(key, 'name')) {
      throw new Error(MISSING_KEY_PROPERTY('name'))
    }

    const name = key.name;
    src = name;

    if (hasOwn.call(key, 'weight')) {
      weight = key.weight;

      if (weight <= 0) {
        throw new Error(INVALID_KEY_WEIGHT_VALUE(name))
      }
    }

    path = createKeyPath(name);
    id = createKeyId(name);
  }

  return { path, id, weight, src }
}

function createKeyPath(key) {
  return isArray(key) ? key : key.split('.')
}

function createKeyId(key) {
  return isArray(key) ? key.join('.') : key
}

function get(obj, path) {
  let list = [];
  let arr = false;

  const deepGet = (obj, path, index) => {
    if (!isDefined(obj)) {
      return
    }
    if (!path[index]) {
      // If there's no path left, we've arrived at the object we care about.
      list.push(obj);
    } else {
      let key = path[index];

      const value = obj[key];

      if (!isDefined(value)) {
        return
      }

      // If we're at the last value in the path, and if it's a string/number/bool,
      // add it to the list
      if (
        index === path.length - 1 &&
        (isString(value) || isNumber(value) || isBoolean(value))
      ) {
        list.push(toString(value));
      } else if (isArray(value)) {
        arr = true;
        // Search each item in the array.
        for (let i = 0, len = value.length; i < len; i += 1) {
          deepGet(value[i], path, index + 1);
        }
      } else if (path.length) {
        // An object. Recurse further.
        deepGet(value, path, index + 1);
      }
    }
  };

  // Backwards compatibility (since path used to be a string)
  deepGet(obj, isString(path) ? path.split('.') : path, 0);

  return arr ? list : list[0]
}

const MatchOptions = {
  // Whether the matches should be included in the result set. When `true`, each record in the result
  // set will include the indices of the matched characters.
  // These can consequently be used for highlighting purposes.
  includeMatches: false,
  // When `true`, the matching function will continue to the end of a search pattern even if
  // a perfect match has already been located in the string.
  findAllMatches: false,
  // Minimum number of characters that must be matched before a result is considered a match
  minMatchCharLength: 1
};

const BasicOptions = {
  // When `true`, the algorithm continues searching to the end of the input even if a perfect
  // match is found before the end of the same input.
  isCaseSensitive: false,
  // When true, the matching function will continue to the end of a search pattern even if
  includeScore: false,
  // List of properties that will be searched. This also supports nested properties.
  keys: [],
  // Whether to sort the result list, by score
  shouldSort: true,
  // Default sort function: sort by ascending score, ascending index
  sortFn: (a, b) =>
    a.score === b.score ? (a.idx < b.idx ? -1 : 1) : a.score < b.score ? -1 : 1
};

const FuzzyOptions = {
  // Approximately where in the text is the pattern expected to be found?
  location: 0,
  // At what point does the match algorithm give up. A threshold of '0.0' requires a perfect match
  // (of both letters and location), a threshold of '1.0' would match anything.
  threshold: 0.6,
  // Determines how close the match must be to the fuzzy location (specified above).
  // An exact letter match which is 'distance' characters away from the fuzzy location
  // would score as a complete mismatch. A distance of '0' requires the match be at
  // the exact location specified, a threshold of '1000' would require a perfect match
  // to be within 800 characters of the fuzzy location to be found using a 0.8 threshold.
  distance: 100
};

const AdvancedOptions = {
  // When `true`, it enables the use of unix-like search commands
  useExtendedSearch: false,
  // The get function to use when fetching an object's properties.
  // The default will search nested paths *ie foo.bar.baz*
  getFn: get,
  // When `true`, search will ignore `location` and `distance`, so it won't matter
  // where in the string the pattern appears.
  // More info: https://fusejs.io/concepts/scoring-theory.html#fuzziness-score
  ignoreLocation: false,
  // When `true`, the calculation for the relevance score (used for sorting) will
  // ignore the field-length norm.
  // More info: https://fusejs.io/concepts/scoring-theory.html#field-length-norm
  ignoreFieldNorm: false,
  // The weight to determine how much field length norm effects scoring.
  fieldNormWeight: 1
};

var Config = {
  ...BasicOptions,
  ...MatchOptions,
  ...FuzzyOptions,
  ...AdvancedOptions
};

const SPACE = /[^ ]+/g;

// Field-length norm: the shorter the field, the higher the weight.
// Set to 3 decimals to reduce index size.
function norm(weight = 1, mantissa = 3) {
  const cache = new Map();
  const m = Math.pow(10, mantissa);

  return {
    get(value) {
      const numTokens = value.match(SPACE).length;

      if (cache.has(numTokens)) {
        return cache.get(numTokens)
      }

      // Default function is 1/sqrt(x), weight makes that variable
      const norm = 1 / Math.pow(numTokens, 0.5 * weight);

      // In place of `toFixed(mantissa)`, for faster computation
      const n = parseFloat(Math.round(norm * m) / m);

      cache.set(numTokens, n);

      return n
    },
    clear() {
      cache.clear();
    }
  }
}

class FuseIndex {
  constructor({
    getFn = Config.getFn,
    fieldNormWeight = Config.fieldNormWeight
  } = {}) {
    this.norm = norm(fieldNormWeight, 3);
    this.getFn = getFn;
    this.isCreated = false;

    this.setIndexRecords();
  }
  setSources(docs = []) {
    this.docs = docs;
  }
  setIndexRecords(records = []) {
    this.records = records;
  }
  setKeys(keys = []) {
    this.keys = keys;
    this._keysMap = {};
    keys.forEach((key, idx) => {
      this._keysMap[key.id] = idx;
    });
  }
  create() {
    if (this.isCreated || !this.docs.length) {
      return
    }

    this.isCreated = true;

    // List is Array<String>
    if (isString(this.docs[0])) {
      this.docs.forEach((doc, docIndex) => {
        this._addString(doc, docIndex);
      });
    } else {
      // List is Array<Object>
      this.docs.forEach((doc, docIndex) => {
        this._addObject(doc, docIndex);
      });
    }

    this.norm.clear();
  }
  // Adds a doc to the end of the index
  add(doc) {
    const idx = this.size();

    if (isString(doc)) {
      this._addString(doc, idx);
    } else {
      this._addObject(doc, idx);
    }
  }
  // Removes the doc at the specified index of the index
  removeAt(idx) {
    this.records.splice(idx, 1);

    // Change ref index of every subsquent doc
    for (let i = idx, len = this.size(); i < len; i += 1) {
      this.records[i].i -= 1;
    }
  }
  getValueForItemAtKeyId(item, keyId) {
    return item[this._keysMap[keyId]]
  }
  size() {
    return this.records.length
  }
  _addString(doc, docIndex) {
    if (!isDefined(doc) || isBlank(doc)) {
      return
    }

    let record = {
      v: doc,
      i: docIndex,
      n: this.norm.get(doc)
    };

    this.records.push(record);
  }
  _addObject(doc, docIndex) {
    let record = { i: docIndex, $: {} };

    // Iterate over every key (i.e, path), and fetch the value at that key
    this.keys.forEach((key, keyIndex) => {
      // console.log(key)
      let value = this.getFn(doc, key.path);

      if (!isDefined(value)) {
        return
      }

      if (isArray(value)) {
        let subRecords = [];
        const stack = [{ nestedArrIndex: -1, value }];

        while (stack.length) {
          const { nestedArrIndex, value } = stack.pop();

          if (!isDefined(value)) {
            continue
          }

          if (isString(value) && !isBlank(value)) {
            let subRecord = {
              v: value,
              i: nestedArrIndex,
              n: this.norm.get(value)
            };

            subRecords.push(subRecord);
          } else if (isArray(value)) {
            value.forEach((item, k) => {
              stack.push({
                nestedArrIndex: k,
                value: item
              });
            });
          } else ;
        }
        record.$[keyIndex] = subRecords;
      } else if (!isBlank(value)) {
        let subRecord = {
          v: value,
          n: this.norm.get(value)
        };

        record.$[keyIndex] = subRecord;
      }
    });

    this.records.push(record);
  }
  toJSON() {
    return {
      keys: this.keys,
      records: this.records
    }
  }
}

function createIndex(
  keys,
  docs,
  { getFn = Config.getFn, fieldNormWeight = Config.fieldNormWeight } = {}
) {
  const myIndex = new FuseIndex({ getFn, fieldNormWeight });
  myIndex.setKeys(keys.map(createKey));
  myIndex.setSources(docs);
  myIndex.create();
  return myIndex
}

function parseIndex(
  data,
  { getFn = Config.getFn, fieldNormWeight = Config.fieldNormWeight } = {}
) {
  const { keys, records } = data;
  const myIndex = new FuseIndex({ getFn, fieldNormWeight });
  myIndex.setKeys(keys);
  myIndex.setIndexRecords(records);
  return myIndex
}

function computeScore$1(
  pattern,
  {
    errors = 0,
    currentLocation = 0,
    expectedLocation = 0,
    distance = Config.distance,
    ignoreLocation = Config.ignoreLocation
  } = {}
) {
  const accuracy = errors / pattern.length;

  if (ignoreLocation) {
    return accuracy
  }

  const proximity = Math.abs(expectedLocation - currentLocation);

  if (!distance) {
    // Dodge divide by zero error.
    return proximity ? 1.0 : accuracy
  }

  return accuracy + proximity / distance
}

function convertMaskToIndices(
  matchmask = [],
  minMatchCharLength = Config.minMatchCharLength
) {
  let indices = [];
  let start = -1;
  let end = -1;
  let i = 0;

  for (let len = matchmask.length; i < len; i += 1) {
    let match = matchmask[i];
    if (match && start === -1) {
      start = i;
    } else if (!match && start !== -1) {
      end = i - 1;
      if (end - start + 1 >= minMatchCharLength) {
        indices.push([start, end]);
      }
      start = -1;
    }
  }

  // (i-1 - start) + 1 => i - start
  if (matchmask[i - 1] && i - start >= minMatchCharLength) {
    indices.push([start, i - 1]);
  }

  return indices
}

// Machine word size
const MAX_BITS = 32;

function search(
  text,
  pattern,
  patternAlphabet,
  {
    location = Config.location,
    distance = Config.distance,
    threshold = Config.threshold,
    findAllMatches = Config.findAllMatches,
    minMatchCharLength = Config.minMatchCharLength,
    includeMatches = Config.includeMatches,
    ignoreLocation = Config.ignoreLocation
  } = {}
) {
  if (pattern.length > MAX_BITS) {
    throw new Error(PATTERN_LENGTH_TOO_LARGE(MAX_BITS))
  }

  const patternLen = pattern.length;
  // Set starting location at beginning text and initialize the alphabet.
  const textLen = text.length;
  // Handle the case when location > text.length
  const expectedLocation = Math.max(0, Math.min(location, textLen));
  // Highest score beyond which we give up.
  let currentThreshold = threshold;
  // Is there a nearby exact match? (speedup)
  let bestLocation = expectedLocation;

  // Performance: only computer matches when the minMatchCharLength > 1
  // OR if `includeMatches` is true.
  const computeMatches = minMatchCharLength > 1 || includeMatches;
  // A mask of the matches, used for building the indices
  const matchMask = computeMatches ? Array(textLen) : [];

  let index;

  // Get all exact matches, here for speed up
  while ((index = text.indexOf(pattern, bestLocation)) > -1) {
    let score = computeScore$1(pattern, {
      currentLocation: index,
      expectedLocation,
      distance,
      ignoreLocation
    });

    currentThreshold = Math.min(score, currentThreshold);
    bestLocation = index + patternLen;

    if (computeMatches) {
      let i = 0;
      while (i < patternLen) {
        matchMask[index + i] = 1;
        i += 1;
      }
    }
  }

  // Reset the best location
  bestLocation = -1;

  let lastBitArr = [];
  let finalScore = 1;
  let binMax = patternLen + textLen;

  const mask = 1 << (patternLen - 1);

  for (let i = 0; i < patternLen; i += 1) {
    // Scan for the best match; each iteration allows for one more error.
    // Run a binary search to determine how far from the match location we can stray
    // at this error level.
    let binMin = 0;
    let binMid = binMax;

    while (binMin < binMid) {
      const score = computeScore$1(pattern, {
        errors: i,
        currentLocation: expectedLocation + binMid,
        expectedLocation,
        distance,
        ignoreLocation
      });

      if (score <= currentThreshold) {
        binMin = binMid;
      } else {
        binMax = binMid;
      }

      binMid = Math.floor((binMax - binMin) / 2 + binMin);
    }

    // Use the result from this iteration as the maximum for the next.
    binMax = binMid;

    let start = Math.max(1, expectedLocation - binMid + 1);
    let finish = findAllMatches
      ? textLen
      : Math.min(expectedLocation + binMid, textLen) + patternLen;

    // Initialize the bit array
    let bitArr = Array(finish + 2);

    bitArr[finish + 1] = (1 << i) - 1;

    for (let j = finish; j >= start; j -= 1) {
      let currentLocation = j - 1;
      let charMatch = patternAlphabet[text.charAt(currentLocation)];

      if (computeMatches) {
        // Speed up: quick bool to int conversion (i.e, `charMatch ? 1 : 0`)
        matchMask[currentLocation] = +!!charMatch;
      }

      // First pass: exact match
      bitArr[j] = ((bitArr[j + 1] << 1) | 1) & charMatch;

      // Subsequent passes: fuzzy match
      if (i) {
        bitArr[j] |=
          ((lastBitArr[j + 1] | lastBitArr[j]) << 1) | 1 | lastBitArr[j + 1];
      }

      if (bitArr[j] & mask) {
        finalScore = computeScore$1(pattern, {
          errors: i,
          currentLocation,
          expectedLocation,
          distance,
          ignoreLocation
        });

        // This match will almost certainly be better than any existing match.
        // But check anyway.
        if (finalScore <= currentThreshold) {
          // Indeed it is
          currentThreshold = finalScore;
          bestLocation = currentLocation;

          // Already passed `loc`, downhill from here on in.
          if (bestLocation <= expectedLocation) {
            break
          }

          // When passing `bestLocation`, don't exceed our current distance from `expectedLocation`.
          start = Math.max(1, 2 * expectedLocation - bestLocation);
        }
      }
    }

    // No hope for a (better) match at greater error levels.
    const score = computeScore$1(pattern, {
      errors: i + 1,
      currentLocation: expectedLocation,
      expectedLocation,
      distance,
      ignoreLocation
    });

    if (score > currentThreshold) {
      break
    }

    lastBitArr = bitArr;
  }

  const result = {
    isMatch: bestLocation >= 0,
    // Count exact matches (those with a score of 0) to be "almost" exact
    score: Math.max(0.001, finalScore)
  };

  if (computeMatches) {
    const indices = convertMaskToIndices(matchMask, minMatchCharLength);
    if (!indices.length) {
      result.isMatch = false;
    } else if (includeMatches) {
      result.indices = indices;
    }
  }

  return result
}

function createPatternAlphabet(pattern) {
  let mask = {};

  for (let i = 0, len = pattern.length; i < len; i += 1) {
    const char = pattern.charAt(i);
    mask[char] = (mask[char] || 0) | (1 << (len - i - 1));
  }

  return mask
}

class BitapSearch {
  constructor(
    pattern,
    {
      location = Config.location,
      threshold = Config.threshold,
      distance = Config.distance,
      includeMatches = Config.includeMatches,
      findAllMatches = Config.findAllMatches,
      minMatchCharLength = Config.minMatchCharLength,
      isCaseSensitive = Config.isCaseSensitive,
      ignoreLocation = Config.ignoreLocation
    } = {}
  ) {
    this.options = {
      location,
      threshold,
      distance,
      includeMatches,
      findAllMatches,
      minMatchCharLength,
      isCaseSensitive,
      ignoreLocation
    };

    this.pattern = isCaseSensitive ? pattern : pattern.toLowerCase();

    this.chunks = [];

    if (!this.pattern.length) {
      return
    }

    const addChunk = (pattern, startIndex) => {
      this.chunks.push({
        pattern,
        alphabet: createPatternAlphabet(pattern),
        startIndex
      });
    };

    const len = this.pattern.length;

    if (len > MAX_BITS) {
      let i = 0;
      const remainder = len % MAX_BITS;
      const end = len - remainder;

      while (i < end) {
        addChunk(this.pattern.substr(i, MAX_BITS), i);
        i += MAX_BITS;
      }

      if (remainder) {
        const startIndex = len - MAX_BITS;
        addChunk(this.pattern.substr(startIndex), startIndex);
      }
    } else {
      addChunk(this.pattern, 0);
    }
  }

  searchIn(text) {
    const { isCaseSensitive, includeMatches } = this.options;

    if (!isCaseSensitive) {
      text = text.toLowerCase();
    }

    // Exact match
    if (this.pattern === text) {
      let result = {
        isMatch: true,
        score: 0
      };

      if (includeMatches) {
        result.indices = [[0, text.length - 1]];
      }

      return result
    }

    // Otherwise, use Bitap algorithm
    const {
      location,
      distance,
      threshold,
      findAllMatches,
      minMatchCharLength,
      ignoreLocation
    } = this.options;

    let allIndices = [];
    let totalScore = 0;
    let hasMatches = false;

    this.chunks.forEach(({ pattern, alphabet, startIndex }) => {
      const { isMatch, score, indices } = search(text, pattern, alphabet, {
        location: location + startIndex,
        distance,
        threshold,
        findAllMatches,
        minMatchCharLength,
        includeMatches,
        ignoreLocation
      });

      if (isMatch) {
        hasMatches = true;
      }

      totalScore += score;

      if (isMatch && indices) {
        allIndices = [...allIndices, ...indices];
      }
    });

    let result = {
      isMatch: hasMatches,
      score: hasMatches ? totalScore / this.chunks.length : 1
    };

    if (hasMatches && includeMatches) {
      result.indices = allIndices;
    }

    return result
  }
}

const registeredSearchers = [];

function createSearcher(pattern, options) {
  for (let i = 0, len = registeredSearchers.length; i < len; i += 1) {
    let searcherClass = registeredSearchers[i];
    if (searcherClass.condition(pattern, options)) {
      return new searcherClass(pattern, options)
    }
  }

  return new BitapSearch(pattern, options)
}

const LogicalOperator = {
  AND: '$and',
  OR: '$or'
};

const KeyType = {
  PATH: '$path',
  PATTERN: '$val'
};

const isExpression = (query) =>
  !!(query[LogicalOperator.AND] || query[LogicalOperator.OR]);

const isPath = (query) => !!query[KeyType.PATH];

const isLeaf = (query) =>
  !isArray(query) && isObject(query) && !isExpression(query);

const convertToExplicit = (query) => ({
  [LogicalOperator.AND]: Object.keys(query).map((key) => ({
    [key]: query[key]
  }))
});

// When `auto` is `true`, the parse function will infer and initialize and add
// the appropriate `Searcher` instance
function parse(query, options, { auto = true } = {}) {
  const next = (query) => {
    let keys = Object.keys(query);

    const isQueryPath = isPath(query);

    if (!isQueryPath && keys.length > 1 && !isExpression(query)) {
      return next(convertToExplicit(query))
    }

    if (isLeaf(query)) {
      const key = isQueryPath ? query[KeyType.PATH] : keys[0];

      const pattern = isQueryPath ? query[KeyType.PATTERN] : query[key];

      if (!isString(pattern)) {
        throw new Error(LOGICAL_SEARCH_INVALID_QUERY_FOR_KEY(key))
      }

      const obj = {
        keyId: createKeyId(key),
        pattern
      };

      if (auto) {
        obj.searcher = createSearcher(pattern, options);
      }

      return obj
    }

    let node = {
      children: [],
      operator: keys[0]
    };

    keys.forEach((key) => {
      const value = query[key];

      if (isArray(value)) {
        value.forEach((item) => {
          node.children.push(next(item));
        });
      }
    });

    return node
  };

  if (!isExpression(query)) {
    query = convertToExplicit(query);
  }

  return next(query)
}

// Practical scoring function
function computeScore(
  results,
  { ignoreFieldNorm = Config.ignoreFieldNorm }
) {
  results.forEach((result) => {
    let totalScore = 1;

    result.matches.forEach(({ key, norm, score }) => {
      const weight = key ? key.weight : null;

      totalScore *= Math.pow(
        score === 0 && weight ? Number.EPSILON : score,
        (weight || 1) * (ignoreFieldNorm ? 1 : norm)
      );
    });

    result.score = totalScore;
  });
}

function transformMatches(result, data) {
  const matches = result.matches;
  data.matches = [];

  if (!isDefined(matches)) {
    return
  }

  matches.forEach((match) => {
    if (!isDefined(match.indices) || !match.indices.length) {
      return
    }

    const { indices, value } = match;

    let obj = {
      indices,
      value
    };

    if (match.key) {
      obj.key = match.key.src;
    }

    if (match.idx > -1) {
      obj.refIndex = match.idx;
    }

    data.matches.push(obj);
  });
}

function transformScore(result, data) {
  data.score = result.score;
}

function format(
  results,
  docs,
  {
    includeMatches = Config.includeMatches,
    includeScore = Config.includeScore
  } = {}
) {
  const transformers = [];

  if (includeMatches) transformers.push(transformMatches);
  if (includeScore) transformers.push(transformScore);

  return results.map((result) => {
    const { idx } = result;

    const data = {
      item: docs[idx],
      refIndex: idx
    };

    if (transformers.length) {
      transformers.forEach((transformer) => {
        transformer(result, data);
      });
    }

    return data
  })
}

class Fuse {
  constructor(docs, options = {}, index) {
    this.options = { ...Config, ...options };

    if (
      this.options.useExtendedSearch &&
      !false
    ) {
      throw new Error(EXTENDED_SEARCH_UNAVAILABLE)
    }

    this._keyStore = new KeyStore(this.options.keys);

    this.setCollection(docs, index);
  }

  setCollection(docs, index) {
    this._docs = docs;

    if (index && !(index instanceof FuseIndex)) {
      throw new Error(INCORRECT_INDEX_TYPE)
    }

    this._myIndex =
      index ||
      createIndex(this.options.keys, this._docs, {
        getFn: this.options.getFn,
        fieldNormWeight: this.options.fieldNormWeight
      });
  }

  add(doc) {
    if (!isDefined(doc)) {
      return
    }

    this._docs.push(doc);
    this._myIndex.add(doc);
  }

  remove(predicate = (/* doc, idx */) => false) {
    const results = [];

    for (let i = 0, len = this._docs.length; i < len; i += 1) {
      const doc = this._docs[i];
      if (predicate(doc, i)) {
        this.removeAt(i);
        i -= 1;
        len -= 1;

        results.push(doc);
      }
    }

    return results
  }

  removeAt(idx) {
    this._docs.splice(idx, 1);
    this._myIndex.removeAt(idx);
  }

  getIndex() {
    return this._myIndex
  }

  search(query, { limit = -1 } = {}) {
    const {
      includeMatches,
      includeScore,
      shouldSort,
      sortFn,
      ignoreFieldNorm
    } = this.options;

    let results = isString(query)
      ? isString(this._docs[0])
        ? this._searchStringList(query)
        : this._searchObjectList(query)
      : this._searchLogical(query);

    computeScore(results, { ignoreFieldNorm });

    if (shouldSort) {
      results.sort(sortFn);
    }

    if (isNumber(limit) && limit > -1) {
      results = results.slice(0, limit);
    }

    return format(results, this._docs, {
      includeMatches,
      includeScore
    })
  }

  _searchStringList(query) {
    const searcher = createSearcher(query, this.options);
    const { records } = this._myIndex;
    const results = [];

    // Iterate over every string in the index
    records.forEach(({ v: text, i: idx, n: norm }) => {
      if (!isDefined(text)) {
        return
      }

      const { isMatch, score, indices } = searcher.searchIn(text);

      if (isMatch) {
        results.push({
          item: text,
          idx,
          matches: [{ score, value: text, norm, indices }]
        });
      }
    });

    return results
  }

  _searchLogical(query) {
    {
      throw new Error(LOGICAL_SEARCH_UNAVAILABLE)
    }
  }

  _searchObjectList(query) {
    const searcher = createSearcher(query, this.options);
    const { keys, records } = this._myIndex;
    const results = [];

    // List is Array<Object>
    records.forEach(({ $: item, i: idx }) => {
      if (!isDefined(item)) {
        return
      }

      let matches = [];

      // Iterate over every key (i.e, path), and fetch the value at that key
      keys.forEach((key, keyIndex) => {
        matches.push(
          ...this._findMatches({
            key,
            value: item[keyIndex],
            searcher
          })
        );
      });

      if (matches.length) {
        results.push({
          idx,
          item,
          matches
        });
      }
    });

    return results
  }
  _findMatches({ key, value, searcher }) {
    if (!isDefined(value)) {
      return []
    }

    let matches = [];

    if (isArray(value)) {
      value.forEach(({ v: text, i: idx, n: norm }) => {
        if (!isDefined(text)) {
          return
        }

        const { isMatch, score, indices } = searcher.searchIn(text);

        if (isMatch) {
          matches.push({
            score,
            key,
            value: text,
            idx,
            norm,
            indices
          });
        }
      });
    } else {
      const { v: text, n: norm } = value;

      const { isMatch, score, indices } = searcher.searchIn(text);

      if (isMatch) {
        matches.push({ score, key, value: text, norm, indices });
      }
    }

    return matches
  }
}

Fuse.version = '6.5.3';
Fuse.createIndex = createIndex;
Fuse.parseIndex = parseIndex;
Fuse.config = Config;

{
  Fuse.parseQuery = parse;
}

/* global HTMLElement, customElements */


const queryInputField = () => domElement$1(`
  <input type="text" class="fuzzy-combo-query-field form-control" />
`);

const searchResultList = () => domElement$1(`
  <ul class="list-group fuzzy-combo-result-list hidden card"></ul>
`);

const searchResultItem = (label, refIndex) => domElement$1(`
  <li class="list-group-item">
    <fuzzy-combo-result-item>
      <button data-ref-index="${refIndex}" class="btn btn-link" type="button">${label}</button>
    </fuzzy-combo-result-item>
  </li>
`);

const clearButton = (fieldLabel) => domElement$1(`<button class="fuzzy-combo-clear-option btn btn-link" type="button" aria-label="clear ${fieldLabel}">X</button>`);

class FuzzyComboBox extends HTMLElement {
  constructor () {
    super();

    this.queryInputField = queryInputField();
    this.searchResultList = searchResultList();
    this.suggestionLimit = this.hasAttribute('suggestion-limit') ? parseInt(this.getAttribute('suggestion-limit')) : 10;
  }

  connectedCallback () {
    this.select?.classList.add('hidden'); // hide original select
    if (this.select?.hasAttribute('tabindex')) {
      this.queryInputField?.setAttribute('tabindex', this.select.getAttribute('tabindex'));
      this.select.setAttribute('tabindex', null);
    }
    this.select.after(this.queryInputField); // replace with text field for searching
    this.queryInputField.after(this.searchResultList); // append searc result list, to pick a list entry

    this.fuse = new Fuse(this.options, { keys: ['label'] });

    this.searchListener = debounce(() => this.search(), 50);
    this.queryInputField.addEventListener('input', this.searchListener);
    this.queryInputField.addEventListener('click', this.searchListener);

    const preselectedOption = this.querySelector('option[selected]');

    if (preselectedOption) {
      this.queryInputField.value = preselectedOption.label;
    }

    this.selectionListener = (ev) => this.selectOption(ev.detail.refIndex);
    this.addEventListener('fuzzy-search:selected', this.selectionListener);

    this.keyboardListener = ev => {
      if (ev.keyCode === 40) {
        this.enterListByKeyboard();
      } else if (ev.keyCode === 27) {
        this.clearSearchResults();
        this.reactivateQueryFocus();
      }
    };

    this.regainFocusListener = ev => this.reactivateQueryFocus();
    this.addEventListener('fuzzy-search-item:focusout', this.regainFocusListener);

    this.cancelationListener = ev => {
      this.clearSearchResults();
      this.reactivateQueryFocus();
    };

    this.addEventListener('fuzzy-search-item:cancel', this.cancelationListener);

    this.queryInputField?.addEventListener('keydown', this.keyboardListener);

    this.focusOutListener = ev => this.focusOut(ev);
    this.addEventListener('focusout', this.focusOutListener);

    const nullOption = this.select?.querySelector('option[value=""]');
    if (nullOption) {
      this.clearOptionButton = clearButton('fubar');
      this.appendChild(this.clearOptionButton);

      this.clearButtonListener = e => {
        e.preventDefault();
        this.deselectAll();
        this.selectOption(nullOption);
      };
      this.clearOptionButton.addEventListener('click', this.clearButtonListener);
    }
  }

  disconnectedCallback () {
    this.queryInputField?.removeEventListener('input', this.searchListener);
    this.queryInputField?.removeEventListener('focus', this.searchListener);
    this.queryInputField?.removeEventListener('keydown', this.keyboardListener);
    this.removeEventListener('focusout', this.focusOutListener);
    this.removeEventListener('fuzzy-search:selected', this.selectionListener);
    this.removeEventListener('fuzzy-search-item:focusout', this.regainFocusListener);
    this.clearOptionButton?.removeEventListener('click', this.clearButtonListener);
  }

  reactivateQueryFocus () {
    if (this.queryInputField) {
      const end = this.queryInputField.value.length;
      setTimeout(() => { this.queryInputField.setSelectionRange(end, end); }, 10);
      this.queryInputField.focus();
    }
  }

  focusOut (ev) {
    if (this.contains(ev.relatedTarget || ev.explicitOriginalTarget || ev.target)) {
      // click has been focussed to a sub-element, which we don't handle
      return
    }

    this.clearSearchResults();
  }

  enterListByKeyboard () {
    if (this.searchResultList.classList.contains('open')) {
      const firstItem = this.searchResultList.querySelector('fuzzy-combo-result-item');
      firstItem?.activate();
    } else {
      this.search();
    }
  }

  selectOption (itemRef) {
    const option = this.select.item(itemRef);

    this.deselectAll(); // multi-selected is not supported

    if (option) {
      option.selected = true;
      option.setAttribute('selected', true);
      this.queryInputField.value = option.label;
    }

    this.clearSearchResults();
    this.queryInputField.focus();
  }

  get select () {
    if (this._select) {
      return this._select
    }

    this._select = this.querySelector('select');

    if (!this._select) {
      console.log('FuzzyComboBoxError: no select field found');
    }

    return this._select
  }

  get options () {
    if (this._options) {
      return this._options
    }

    this._options = Array.from(this.select?.options);

    if (!this._options) {
      console.log('FuzzyComboBoxError: no options fields found');
    }

    return this._options
  }

  search () {
    const searchTerm = this.queryInputField?.value;
    this.clearSearchResults();
    if (searchTerm) {
      const results = this.fuse.search(searchTerm);

      results
        .map(r => searchResultItem(this.options[r.refIndex]?.label, r.refIndex))
        .slice(0, this.suggestionLimit) // limit to max num of list items
        .forEach(item => this.searchResultList.appendChild(item));

      this.showSearchResults();
    } else {
      const allOptions = this.options?.map((r, i) => r.label && searchResultItem(r.label, i)).filter(x => !!x);
      allOptions.forEach(item => this.searchResultList.appendChild(item));

      this.showSearchResults();
    }
  }

  clearSearchResults () {
    this.searchResultList.classList.remove('open');
    this.searchResultList.classList.add('hidden');
    this.searchResultList.innerHTML = '';
  }

  showSearchResults () {
    this.searchResultList.classList.remove('hidden');
    this.searchResultList.classList.add('open');
  }

  deselectAll () {
    this.options?.forEach(o => {
      o.removeAttribute('selected');
      o.selected = false;
    });
    this.queryInputField.value = '';
  }
}

customElements.define('fuzzy-combo-box', FuzzyComboBox);

/* global HTMLElement, customElements, CustomEvent */

class FuzzyComboResultItem extends HTMLElement {
  connectedCallback () {
    this.button = this.querySelector('button');

    this.selectListener = ev => this.select();
    this.button?.addEventListener('click', this.selectListener);

    this.keyboardNavListener = ev => {
      if (ev.keyCode === 40 && this.nextItem) { // on down
        this.deactivate();
        this.nextItem?.activate();
      } else if (ev.keyCode === 38 && this.previousItem) { // on up
        this.deactivate();
        this.previousItem?.activate();
      } else if (ev.keyCode === 38) { // on up on first list item
        this.deactivate();
        this.leaveKeyboardFocus();
      } else if (ev.keyCode === 27) { // on esc
        this.cancelComboBox();
      } else if (ev.keyCode === 9 && !this.previousItem) { // back tab on first item
        ev.preventDefault();
        this.cancelComboBox();
      }
    };

    this.button?.addEventListener('keydown', this.keyboardNavListener);

    this.parentElement.addEventListener('click', this.selectListener);
  }

  disconnectedCallback () {
    this.button?.removeEventListener('click', this.selectListener);
    this.button?.removeEventListener('keydown', this.keyboardNavListener);
    this.parentElement.removeEventListener('focusout', this.focusoutListener);
  }

  get nextItem () {
    if (this._nextItem) {
      return this._nextItem
    }
    this._nextItem = this.parentElement.nextElementSibling?.querySelector('fuzzy-combo-result-item');
    return this._nextItem
  }

  get previousItem () {
    if (this._previousItem) {
      return this._previousItem
    }
    this._previousItem = this.parentElement.previousElementSibling?.querySelector('fuzzy-combo-result-item');
    return this._previousItem
  }

  select () {
    const ev = new CustomEvent('fuzzy-search:selected', { detail: { refIndex: parseInt(this.refIndex) }, bubbles: true });
    this.dispatchEvent(ev);
  }

  leaveKeyboardFocus () {
    const ev = new CustomEvent('fuzzy-search-item:focusout', { bubbles: true });
    this.dispatchEvent(ev);
  }

  cancelComboBox () {
    const ev = new CustomEvent('fuzzy-search-item:cancel', { bubbles: true });
    this.dispatchEvent(ev);
  }

  get refIndex () {
    return this.button?.dataset.refIndex ? this.button.dataset.refIndex : null
  }

  activate () {
    this.button.focus();
    this.classList.add('focussed');
  }

  deactivate () {
    this.button.blur();
    this.classList.remove('focussed');
  }
}

customElements.define('fuzzy-combo-result-item', FuzzyComboResultItem);

/* global HTMLElement, customElements */


const closeButton = () => domElement$1(`
  <div class="d-flex">
    <button class="btn-close me-2 m-auto" type="button" aria-label="Close"></button>
  </div>
`);

class ToastMessage extends HTMLElement {
  constructor () {
    super();
    this.closeButton = closeButton();
  }

  connectedCallback () {
    this.prepend(this.closeButton);

    this.body = this.querySelector('.toast-body');
    this.flexContainer = this.querySelector('.d-flex');
    this.flexContainer.prepend(this.body);

    // the fade class is expected to be rendered, when the component is displayed.
    // So there is no obserer or similar in place to check this.
    if (this.classList.contains('fade')) {
      setTimeout(() => {
        this.fade();
      }, this.fadeAfter);
    }

    this.clickListener = ev => this.fade();
    this.closeButton?.addEventListener('click', this.clickListener);
  }

  disconnectedCallback () {
    this.closeButton?.removeEventListener('click', this.clickListener);
  }

  fade () {
    this.classList.add('fade-now');
    setTimeout(() => { this.remove(); }, 1000);
  }

  get fadeAfter () {
    if (this._fadeAfter) { return this._fadeAfter }

    this._fadeAfter = this.hasAttribute('fade-after') ? parseInt(this.getAttribute('fade-after')) : 1500;

    return this._fadeAfter
  }
}
customElements.define('toast-message', ToastMessage);

/* global HTMLElement, customElements */

class ExpandingForm extends HTMLElement {
  connectedCallback () {
    this.expandingFormButton = this.querySelector('.expand-form-button');
    this.form = this.querySelector('.expandable-form');

    this.expandListener = ev => this.toggleExpand();
    this.expandingFormButton?.addEventListener('click', this.expandListener);
  }

  disconnectedCallback () {
    this.expandingFormButton?.removeEventListener('click', this.expandListener);
  }

  toggleExpand () {
    this.form?.classList.toggle('hidden');
  }
}
customElements.define('expanding-form', ExpandingForm);

/* global HTMLElement, customElements, IntersectionObserver */
class ScrollFoam extends HTMLElement {
  connectedCallback () {
    this.mainElement = this.querySelector('main');

    if (typeof IntersectionObserver !== 'undefined') {
      this.observer = new IntersectionObserver((entries) => {
      // isIntersecting is true when element and viewport are overlapping
      // isIntersecting is false when element and viewport don't overlap
        if (entries[0].isIntersecting === true) {
          this.classList.remove('squeeze-out');
          this.classList.add('squeeze');
        } else {
          this.classList.add('squeeze-out');
        }
      }, { threshold: [0] });
    }

    this.observer.observe(document.querySelector('footer'));
  }

  disconnectedCallback () {
    this.observer?.disconnect();
  }
}

customElements.define('scroll-foam', ScrollFoam);

/* global HTMLElement, customElements, document */

const baseSvg = () => {
  const svg = document.createElementNS('http://www.w3.org/2000/svg', 'svg');
  svg.classList.add('connection-visualizer-canvas');
  return svg
};

const svgLine = (x1, y1, x2, y2, { stroke = '#000000', strokeWidth = '1px' } = {}) => {
  const line = document.createElementNS('http://www.w3.org/2000/svg', 'line');
  line.setAttribute('stroke-width', strokeWidth);
  line.setAttribute('stroke', stroke);
  line.setAttribute('stroke-linecap', 'round');

  line.setAttribute('x1', x1);
  line.setAttribute('y1', y1);
  line.setAttribute('x2', x2);
  line.setAttribute('y2', y2);

  return line
};

class ConnectionEntity {
  constructor (entity, { container = null } = {}) {
    this.entity = entity;
    this.container = container;
    this.lines = [];

    this.rect = this.entity.getBoundingClientRect();
  }

  get id () {
    if (this._id) { return this._id }

    this._id = this.entity?.hasAttribute('entity-id') ? this.entity.getAttribute('entity-id') : null;

    return this._id
  }

  get offset () {
    const offset = {
      top: this.rect.top + window.scrollY - this.containerOffset.top,
      right: this.rect.right + window.scrollX - this.containerOffset.left,
      bottom: this.rect.bottom + window.scrollY - this.containerOffset.top,
      left: this.rect.left + window.scrollX - this.containerOffset.left
    };

    return offset
  }

  get containerOffset () {
    return this.container
      ? this.container.offset
      : {
          top: 0,
          right: 0,
          bottom: 0,
          left: 0
        }
  }

  get width () {
    return this.rect.width
  }

  get height () {
    return this.rect.height
  }

  refreshDimensions () {
    this.rect = this.entity.getBoundingClientRect();
  }
}

class ConnectionVisualizer extends HTMLElement {
  constructor () {
    super();
    this.svg = baseSvg();
    this.gap = this.hasAttribute('gap') ? parseInt(this.getAttribute('gap')) : undefined;
    this.stroke = this.hasAttribute('stroke') ? this.getAttribute('stroke') : undefined;
    this.strokeWidth = this.hasAttribute('stroke-width') ? this.getAttribute('stroke-width') : undefined;
  }

  connectedCallback () {
    this.appendChild(this.svg);
    this.container = new ConnectionEntity(this.svg);

    this.sources = toArray(this.querySelectorAll('connection-source')).map(s => new ConnectionEntity(s, { container: this.container }));

    this.sources.forEach(s => {
      s.hoverListener = e => {
        this.drawLines(s);
      };

      s.hoverOutListener = e => {
        this.removeLines();
      };

      s.entity.addEventListener('mouseover', s.hoverListener);
      s.entity.addEventListener('mouseout', s.hoverOutListener);
    });
  }

  disconnectedCallback () {
    this.sources.forEach(s => {
      s.entity.removeEventListener('mouseover', s.hoverListener);
      s.entity.removeEventListener('mouseout', s.hoverOutListener);
    });
  }

  drawLines (source) {
    source.container.refreshDimensions();
    source.refreshDimensions();
    this.removeLines();
    const targets = toArray(this.querySelectorAll('connection-target'))
      .map(t => new ConnectionEntity(t, { container: this.container }))
      .filter(t => t.id === source.id);

    window.requestAnimationFrame(() => {
      targets.forEach(t => {
        t.refreshDimensions();
        const line = svgLine(...connectionCoordinates(source, t, { gap: this.gap }), { stroke: this.stroke, strokeWidth: this.strokeWidth });
        this.svg.appendChild(line);
        source.lines.push(line);
      });
    });
  }

  removeLines () {
    this.sources.forEach(s => {
      s.lines.forEach(l => l.remove());
    });
  }
}

customElements.define('connection-visualizer', ConnectionVisualizer);

function toArray (nodeList) {
  return Array.prototype.slice.call(nodeList)
}

function connectionCoordinates (source, target, { gap = 0 } = {}) {
  let x1, x2;

  const y1 = source.offset.top + source.height / 2;
  const y2 = target.offset.top + target.height / 2;

  let sourceIsLeft = false;
  let sourceIsRight = false;

  if (source.offset.right < target.offset.left) {
    sourceIsLeft = true;
    x1 = source.offset.right; // use right side of source as x-axis anchor
    x2 = target.offset.left; // use left side of target as x-axis anchor
  } else if (source.offset.left < target.offset.right) { // source overlaps with target on x-axis
    x1 = source.offset.left + source.width / 2; // use source middle as x-anchor
    x2 = target.offset.left + target.width / 2; // use target middle as x-anchor
  } else { // source is right of target
    sourceIsRight = true;
    x1 = source.offset.left; // use left side of source as x-axis anchor
    x2 = target.offset.right; // use right side of target as x-axis anchor
  }

  if (sourceIsLeft) {
    x1 = x1 + gap;
    x2 = x2 - gap;
  } else if (sourceIsRight) {
    x1 = x1 - gap;
    x2 = x2 + gap;
  }

  return [x1, y1, x2, y2]
}

/* global customElements, HTMLElement */

const label = (classes, activates) => domElement$1(`
  <label class="${classes}" for="${activates}"></label>
`);

const checkbox = (id) => domElement$1(`
  <input class="on-demand-state-checkbox" type="checkbox" name="${id}" id="${id}">
`);

class OnDemandActions extends HTMLElement {
  constructor () {
    super();

    const uuid = Math.ceil(Math.random() * 100000);
    const modelStateId = `model-state-${uuid}`;
    this.modelState = checkbox(modelStateId);
    this.openLabel = label('on-demand-open-label', modelStateId);
    this.backdropLabel = label('on-demand-backdrop-label hidden', modelStateId);
  }

  connectedCallback () {
    this.prepend(this.modelState);
    this.appendChild(this.openLabel);
    this.appendChild(this.backdropLabel);
  }
}

customElements.define('on-demand-actions', OnDemandActions);

/* global customElements, HTMLElement, fetch,  */


const DISABLE_COOKIE_NAME = 'disable-connected-resources';

const liveLabel = () => {
  let i = 0;
  let iteration = 0;
  const element = domElement$1('<button class="btn btn-sm badge badge-warning mb-3">Live</button>');
  const liveIndicationAnimationInterval = setInterval(() => {
    i = (i + 1) % 4;
    iteration += 1;
    element.textContent = `Live${'.'.repeat ? '.'.repeat(i) : ''}`;

    if (iteration >= 3) {
      element.classList.remove('badge-warning');
      element.classList.add('badge-success');
      clearInterval(liveIndicationAnimationInterval);
    }
  }, 500);
  return element
};

const stoppedLabel = () => domElement$1('<button class="btn btn-sm badge badge-danger mb-3">Dead</button>');

const progressBar = () => domElement$1('<connected-resource-progress-wrapper><div class="custom-progress hidden" value="0" max="100"></connected-resource-progress-wrapper>');

const metaWrapper = () => domElement$1('<connected-resource-meta />');

const POLL_INTERVAL = 2500;
class ConnectedResource extends HTMLElement {
  constructor () {
    super();
    this.liveLabelMouseoverListener = () => { this.classList.add('with-hint'); };
    this.liveLabelMouseoutListener = () => { this.classList.remove('with-hint'); };
    this.metaWrapper = metaWrapper();
    this.progressBar = progressBar();
  }

  async connectedCallback () {
    this.before(this.metaWrapper);
    this.metaWrapper.appendChild(this.progressBar);
    this.progress = this.progressBar.querySelector('div');

    this.sourcePath = this.getAttribute('src');
    this.filterDoc = this.hasAttribute('selector')
      ? (doc) => querySelectorFromString(doc, this.getAttribute('selector'))
      : (doc) => doc;

    await this.start({ resetCookie: false });

    if (this.isDisabled) {
      this.stop();
    }
  }

  async updateContentWhenChanged (result) {
    const etag = result.headers.get('etag');
    if (etag) {
      const lastEtag = this.getAttribute('last-etag');
      if (etag === lastEtag) {
        return
      }
      this.progress.classList.remove('hidden');
      this.setAttribute('last-etag', etag);
    }
    // skip when user is entering content in the updatable area.
    // Wait for next iteration and check again.

    if (!this.contains(document.activeElement)) {
      const text = await result.text();
      const htmlString = this.filterDoc(text);
      this.innerHTML = htmlString;
      this.progress.style.width = '100%';
    }
  }

  async update () {
    if (this.pollForUpdates) {
      try {
        const result = await fetch(this.sourcePath, { credentials: 'include' });

        if (result.ok) {
          await this.updateContentWhenChanged(result);
        } else {
          this.stop();
        }
      } catch (e) {
        console.log(e);
        this.stop();
        this.progress.classList.remove('hidden');
        this.progress.classList.add('error');
        this.progress.style.width = '100%';
      } finally {
        setTimeout(() => {
          if (!this.progress.classList.contains('error')) {
            this.progress.classList.add('hidden');
            this.progress.style.width = 0;
          }
        }, 1000);
      }

      setTimeout(() => {
        this.update();
      }, POLL_INTERVAL);
    }
  }

  async start ({ resetCookie = true } = {}) {
    if (resetCookie) {
      console.log('resetting cookie!');
      removeDisabledCookie();
    }
    this.progress.classList.add('hidden');
    this.progress.classList.remove('error');
    this.progress.style.width = 0;

    this.pollForUpdates = true;
    this.liveLabel = liveLabel();
    this.metaWrapper.before(this.liveLabel);
    this.liveLabel.addEventListener('click', () => this.stop());
    this.liveLabel.addEventListener('mouseover', this.liveLabelMouseoverListener);
    this.liveLabel.addEventListener('mouseout', this.liveLabelMouseoutListener);
    this.stoppedLabel?.remove();

    setTimeout(async () => {
      await this.update();
    }, POLL_INTERVAL);
  }

  stop () {
    setDisabledCookie();
    if (this.liveLabel) {
      this.liveLabelMouseoutListener();
      this.liveLabel.addEventListener('mouseover', this.liveLabelMouseoverListener);
      this.liveLabel.addEventListener('mouseout', this.liveLabelMouseoutListener);
      this.liveLabel.remove();
    }

    this.pollForUpdates = false;
    this.stoppedLabel = stoppedLabel();
    this.metaWrapper.before(this.stoppedLabel);
    this.stoppedLabel.addEventListener('click', () => this.start());
  }

  disconnectedCallback () {
    if (this.liveLabel) {
      this.liveLabel.addEventListener('mouseover', this.liveLabelMouseoverListener);
      this.liveLabel.addEventListener('mouseout', this.liveLabelMouseoutListener);
      this.liveLabel.remove();
    }
    this.stoppedLabel?.remove();
  }

  get isDisabled () {
    return hasDisabledCookieSet()
  }
}

customElements.define('connected-resource', ConnectedResource);

function setDisabledCookie () {
  document.cookie = `${DISABLE_COOKIE_NAME}=true; path=${window.location.pathname}; SameSite=Lax;`;
}

function removeDisabledCookie () {
  document.cookie = `${DISABLE_COOKIE_NAME}=; path=${window.location.pathname}; Max-Age=0; expires=Thu, 01 Jan 1970 00:00:00 GMT; SameSite=Lax;`;
}

function hasDisabledCookieSet () {
  return document.cookie
    .split(';')
    .map(line => line.trim())
    .findIndex(row => row.startsWith('disable-connected-resources=')) >= 0
}

/* global customElements, HTMLElement  */

class ComparableLists extends HTMLElement {
  connectedCallback () {
    console.log('connected');
    this.reference = this.querySelector('[reference]');
    this.dynamic = this.querySelector('[dynamic]');
    this.scrollContainer = this.dynamic?.querySelector('comparable-list-scroll-container');

    this.scrollListener = debounce(this.adjustDynamicList.bind(this), 50);
    document.addEventListener('scroll', this.scrollListener);

    const dynamicRect = this.dynamic.getBoundingClientRect();
    this.scrollContainer.style['max-height'] = `${limitedToInnerWindowSize(dynamicRect.bottom)}px`;
  }

  disconnectedCallback () {
    document.removeEventListener('scroll', this.scrollListener);
    this.scrollContainer.style['max-height'] = 'auto';
  }

  adjustDynamicList () {
    window.requestAnimationFrame(() => {
      // const outerContainerRect = this.getBoundingClientRect()
      const referenceRect = this.reference.getBoundingClientRect();
      const dynamicRect = this.dynamic.getBoundingClientRect();
      // const scrollContainerRect = this.scrollContainer.getBoundingClientRect()

      const adjustment = (referenceRect.top * -1) + 10;

      const scrollContainerTagetHeight = limitedToInnerWindowSize(dynamicRect.bottom);
      // list is scrolled down and "dynamic" content is below reference content.
      if (referenceRect.top < 0) {
        this.scrollContainer.style.top = `${positivePart(adjustment)}px`;
        this.scrollContainer.scrollTop = 0;
      } else {
        this.scrollContainer.style.top = 0;
        // this.dynamic.style['max-height'] = 'auto'
        this.scrollContainer.scrollTop = 0;
      }

      this.scrollContainer.style['max-height'] = `${scrollContainerTagetHeight}px`;
    });
  }
}

customElements.define('comparable-lists', ComparableLists);

function positivePart (v) {
  return Math.max(0, v)
}

function limitedToInnerWindowSize (val) {
  return Math.min(val, window.innerHeight * 0.9)
}

/* global HTMLElement, customElements */
class BulkSelect extends HTMLElement {
  connectedCallback () {
    this.select = this.querySelector('select');
    this.dependentSelects?.forEach(select => {
      const originalValue = Array.from(select.selectedOptions)[0]?.value;
      if (originalValue) {
        select.setAttribute('data-original-value', originalValue);
      }
    });

    this.selectListener = ev => {
      const selectedOption = Array.from(this.select.selectedOptions)[0];
      if (selectedOption?.value) {
        this.dependentSelects?.forEach(select => {
          const options = Array.from(select.options);
          options?.forEach(option => {
            option.removeAttribute('selected');
            option.selected = false;
          });

          const targetOption = options?.filter(o => o.value === selectedOption.value)[0];
          if (targetOption) {
            targetOption.selected = true;
            targetOption.setAttribute('selected', true);
          }
        });
      } else {
        this.dependentSelects?.forEach(select => {
          if (select.hasAttribute('data-original-value')) {
            const originalValue = select.getAttribute('data-original-value');
            console.log(originalValue);
            const options = Array.from(select.options);
            options?.forEach(option => {
              option.removeAttribute('selected');
              option.selected = false;
            });
            const targetOption = options?.filter(o => o.value === originalValue)[0];

            if (targetOption) {
              console.log(targetOption);
              targetOption.selected = true;
              targetOption.setAttribute('selected', true);
            }
          }
        });
      }
    };
    this.select.addEventListener('change', this.selectListener);
  }

  disconnectedCallback () {
    this.select.removeEventListener('change', this.selectListener);
  }

  get name () {
    return this.getAttribute('name')
  }

  get dependentSelects () {
    // get all select fields with a certain name. As this component should
    // change multiple select-fields for a nested attribute, the name format
    // of the targeted select fields should be in the form `name="property[sub-id]"`
    const selects = document.querySelectorAll(`select[name^="${this.name}["]`);
    return Array.from(selects)
  }
}

customElements.define('bulk-select', BulkSelect);

export { BulkSelect, CombinedSearchForm, ComparableLists, ConnectedResource, ConnectionVisualizer, DateSelect, ExpandingForm, FuzzyComboBox, FuzzyComboResultItem, OnDemandActions, RangeInput, ReactiveFilterSearch, ScrollFoam, ToastMessage };
