/* global HTMLElement, customElements, FormData, fetch, window , CustomEvent */

import { querySelectorFromString, domElement } from '../utils/dom.js'
import { debounce } from '../utils/fp.js'

const inputElement = ({ placeholder } = {}) => domElement(`
  <input type="text" tabindex="1" name="combined-term" min-length="3" ${placeholder ? `placeholder="${placeholder}"` : ''} />
`)

const closeButton = () => domElement('<button class="btn btn-sm close-button" tabindex="2">X</button>')

const submitButton = () => domElement(`
  <button class="btn">&gt;&gt;</button>
`)

export class CombinedSearchForm extends HTMLElement {
  constructor () {
    super()
    this.input = inputElement({ placeholder: this.hasAttribute('label') ? this.getAttribute('label') : '' })
    this.submit = submitButton()
    this.closeButton = closeButton()
  }

  connectedCallback () {
    this.forms = this.querySelectorAll('async-search-form')
    this.formInputs = this.querySelectorAll('async-search-form input[type=text]')
    this.resultView = this.querySelector('combined-search-result-view')
    this.resultView.prepend(this.closeButton)

    if (this.hasSearchHint()) {
      this.resultView.prepend(domElement(`<h5 class="">${this.searchHint}</h5>`))
    }

    this.prepend(this.input)

    this.addEventListener('async-form-data-received', this.activate.bind(this))

    const debouncedSearch = debounce(this.triggerSearch.bind(this), 300)
    this.input.addEventListener('input', debouncedSearch)

    // this.input.addEventListener('focusout', this.deactivate.bind(this))
    this.input.addEventListener('focus', this.activate.bind(this))
    this.closeButton.addEventListener('click', this.deactivate.bind(this))
  }

  triggerSearch () {
    this.formInputs.forEach(input => {
      input.value = this.input.value
    })

    // TODO: use min-length attribute on combined-search-form instead of constant 3
    if (this.input.value && this.input.value.length >= 3) {
      this.forms.forEach(form => form.submit())
    } else {
      // this.deactivate()
      this.forms.forEach(form => form.cleanup())
    }
  }

  activate () {
    this.resultView.classList.add('active')
  }

  deactivate () {
    setTimeout(() =>
      this.resultView.classList.remove('active'), 100)
  }

  get searchHint () {
    return this.getAttribute('search-hint')
  }

  hasSearchHint () {
    return this.hasAttribute('search-hint')
  }
}

const isFQDN = (urlOrPath = '') => urlOrPath.indexOf('://') > 0 || urlOrPath.indexOf('//') === 0

class AsyncSearchForm extends HTMLElement {
  constructor () {
    super()

    this.form = this.querySelector('form')
    this.action = this.form.getAttribute('action')
    this.method = this.form.getAttribute('method')
    this.select = this.getAttribute('select')
  }

  connectedCallback () {
    this.target = this.querySelector('async-search-form-results')

    if (this.method !== 'GET') {
      console.error('async form is only supported for forms using the GET method')
    } else {
      this.form.addEventListener('submit', async ev => {
        ev.preventDefault()
        let searchUrl
        try {
          searchUrl = isFQDN(this.action) ? this.action : this.searchUrl(ev.target)
          const response = await fetch(searchUrl)

          if (!response.ok) {
            if (response.status === 401) {
              console.error('user is logged out, refresh main resource, resulting in a redirect to the Login Provider')
              window.location.reload()
              return
            }

            const message = `could not retrieve data from ${this.action}`
            console.error(message)
            this.renderError(message)
            return
          }

          const results = await response.text()

          const dataReceivedEvent = new CustomEvent('async-form-data-received', { bubbles: true })
          this.dispatchEvent(dataReceivedEvent)

          this.render(results)
        } catch (e) {
          const message = `failed to load resource "${searchUrl.toString()}": ${e.message}`
          console.log(message, e)
          this.renderError(message)
        }
      })
    }
  }

  render (htmlString) {
    if (this.target) {
      this.target.innerHTML = querySelectorFromString(htmlString, this.select)
    } else {
      console.log('result from async form', querySelectorFromString(htmlString, this.select))
    }
  }

  renderError (error) {
    if (this.target) {
      this.target.innerHTML = `<p class="alert alert-danger">${error}</p>`
    }
  }

  searchUrl (form) {
    const queryParams = new FormData(form)
    const query = new URLSearchParams(queryParams).toString()
    const searchUrl = new URL(window.location.href)
    searchUrl.pathname = this.action
    searchUrl.search = query
    return searchUrl
  }

  submit () {
    this.form.requestSubmit()
  }

  cleanup () {
    this.target.innerHTML = ''
  }
}

customElements.define('async-search-form', AsyncSearchForm)
customElements.define('combined-search-form', CombinedSearchForm)
