/* global customElements, HTMLElement, fetch,  */

import { domElement, querySelectorFromString } from '../utils/dom.js'

const DISABLE_COOKIE_NAME = 'disable-connected-resources'

const liveLabel = () => {
  let i = 0
  let iteration = 0
  const element = domElement('<button class="btn btn-sm badge badge-warning mb-3">Live</button>')
  const liveIndicationAnimationInterval = setInterval(() => {
    i = (i + 1) % 4
    iteration += 1
    element.textContent = `Live${'.'.repeat ? '.'.repeat(i) : ''}`

    if (iteration >= 3) {
      element.classList.remove('badge-warning')
      element.classList.add('badge-success')
      clearInterval(liveIndicationAnimationInterval)
    }
  }, 500)
  return element
}

const stoppedLabel = () => domElement('<button class="btn btn-sm badge badge-danger mb-3">Dead</button>')

const progressBar = () => domElement('<connected-resource-progress-wrapper><div class="custom-progress hidden" value="0" max="100"></connected-resource-progress-wrapper>')

const metaWrapper = () => domElement('<connected-resource-meta />')

const POLL_INTERVAL = 2500
export class ConnectedResource extends HTMLElement {
  constructor () {
    super()
    this.liveLabelMouseoverListener = () => { this.classList.add('with-hint') }
    this.liveLabelMouseoutListener = () => { this.classList.remove('with-hint') }
    this.metaWrapper = metaWrapper()
    this.progressBar = progressBar()
  }

  async connectedCallback () {
    this.before(this.metaWrapper)
    this.metaWrapper.appendChild(this.progressBar)
    this.progress = this.progressBar.querySelector('div')

    this.sourcePath = this.getAttribute('src')
    this.filterDoc = this.hasAttribute('selector')
      ? (doc) => querySelectorFromString(doc, this.getAttribute('selector'))
      : (doc) => doc

    await this.start({ resetCookie: false })

    if (this.isDisabled) {
      this.stop()
    }
  }

  async updateContentWhenChanged (result) {
    const etag = result.headers.get('etag')
    if (etag) {
      const lastEtag = this.getAttribute('last-etag')
      if (etag === lastEtag) {
        return
      }
      this.progress.classList.remove('hidden')
      this.setAttribute('last-etag', etag)
    }
    // skip when user is entering content in the updatable area.
    // Wait for next iteration and check again.

    if (!this.contains(document.activeElement)) {
      const text = await result.text()
      const htmlString = this.filterDoc(text)
      this.innerHTML = htmlString
      this.progress.style.width = '100%'
    }
  }

  async update () {
    if (this.pollForUpdates) {
      try {
        const result = await fetch(this.sourcePath, { credentials: 'include' })

        if (result.ok) {
          await this.updateContentWhenChanged(result)
        } else {
          this.stop()
        }
      } catch (e) {
        console.log(e)
        this.stop()
        this.progress.classList.remove('hidden')
        this.progress.classList.add('error')
        this.progress.style.width = '100%'
      } finally {
        setTimeout(() => {
          if (!this.progress.classList.contains('error')) {
            this.progress.classList.add('hidden')
            this.progress.style.width = 0
          }
        }, 1000)
      }

      setTimeout(() => {
        this.update()
      }, POLL_INTERVAL)
    }
  }

  async start ({ resetCookie = true } = {}) {
    if (resetCookie) {
      console.log('resetting cookie!')
      removeDisabledCookie()
    }
    this.progress.classList.add('hidden')
    this.progress.classList.remove('error')
    this.progress.style.width = 0

    this.pollForUpdates = true
    this.liveLabel = liveLabel()
    this.metaWrapper.before(this.liveLabel)
    this.liveLabel.addEventListener('click', () => this.stop())
    this.liveLabel.addEventListener('mouseover', this.liveLabelMouseoverListener)
    this.liveLabel.addEventListener('mouseout', this.liveLabelMouseoutListener)
    this.stoppedLabel?.remove()

    setTimeout(async () => {
      await this.update()
    }, POLL_INTERVAL)
  }

  stop () {
    setDisabledCookie()
    if (this.liveLabel) {
      this.liveLabelMouseoutListener()
      this.liveLabel.addEventListener('mouseover', this.liveLabelMouseoverListener)
      this.liveLabel.addEventListener('mouseout', this.liveLabelMouseoutListener)
      this.liveLabel.remove()
    }

    this.pollForUpdates = false
    this.stoppedLabel = stoppedLabel()
    this.metaWrapper.before(this.stoppedLabel)
    this.stoppedLabel.addEventListener('click', () => this.start())
  }

  disconnectedCallback () {
    if (this.liveLabel) {
      this.liveLabel.addEventListener('mouseover', this.liveLabelMouseoverListener)
      this.liveLabel.addEventListener('mouseout', this.liveLabelMouseoutListener)
      this.liveLabel.remove()
    }
    this.stoppedLabel?.remove()
  }

  get isDisabled () {
    return hasDisabledCookieSet()
  }
}

customElements.define('connected-resource', ConnectedResource)

function setDisabledCookie () {
  document.cookie = `${DISABLE_COOKIE_NAME}=true; path=${window.location.pathname}; SameSite=Lax;`
}

function removeDisabledCookie () {
  document.cookie = `${DISABLE_COOKIE_NAME}=; path=${window.location.pathname}; Max-Age=0; expires=Thu, 01 Jan 1970 00:00:00 GMT; SameSite=Lax;`
}

function hasDisabledCookieSet () {
  return document.cookie
    .split(';')
    .map(line => line.trim())
    .findIndex(row => row.startsWith('disable-connected-resources=')) >= 0
}
