/* global HTMLElement, customElements, document */

const baseSvg = () => {
  const svg = document.createElementNS('http://www.w3.org/2000/svg', 'svg')
  svg.classList.add('connection-visualizer-canvas')
  return svg
}

const svgLine = (x1, y1, x2, y2, { stroke = '#000000', strokeWidth = '1px' } = {}) => {
  const line = document.createElementNS('http://www.w3.org/2000/svg', 'line')
  line.setAttribute('stroke-width', strokeWidth)
  line.setAttribute('stroke', stroke)
  line.setAttribute('stroke-linecap', 'round')

  line.setAttribute('x1', x1)
  line.setAttribute('y1', y1)
  line.setAttribute('x2', x2)
  line.setAttribute('y2', y2)

  return line
}

class ConnectionEntity {
  constructor (entity, { container = null } = {}) {
    this.entity = entity
    this.container = container
    this.lines = []

    this.rect = this.entity.getBoundingClientRect()
  }

  get id () {
    if (this._id) { return this._id }

    this._id = this.entity?.hasAttribute('entity-id') ? this.entity.getAttribute('entity-id') : null

    return this._id
  }

  get offset () {
    const offset = {
      top: this.rect.top + window.scrollY - this.containerOffset.top,
      right: this.rect.right + window.scrollX - this.containerOffset.left,
      bottom: this.rect.bottom + window.scrollY - this.containerOffset.top,
      left: this.rect.left + window.scrollX - this.containerOffset.left
    }

    return offset
  }

  get containerOffset () {
    return this.container
      ? this.container.offset
      : {
          top: 0,
          right: 0,
          bottom: 0,
          left: 0
        }
  }

  get width () {
    return this.rect.width
  }

  get height () {
    return this.rect.height
  }

  refreshDimensions () {
    this.rect = this.entity.getBoundingClientRect()
  }
}

export class ConnectionVisualizer extends HTMLElement {
  constructor () {
    super()
    this.svg = baseSvg()
    this.gap = this.hasAttribute('gap') ? parseInt(this.getAttribute('gap')) : undefined
    this.stroke = this.hasAttribute('stroke') ? this.getAttribute('stroke') : undefined
    this.strokeWidth = this.hasAttribute('stroke-width') ? this.getAttribute('stroke-width') : undefined
  }

  connectedCallback () {
    this.appendChild(this.svg)
    this.container = new ConnectionEntity(this.svg)

    this.sources = toArray(this.querySelectorAll('connection-source')).map(s => new ConnectionEntity(s, { container: this.container }))

    this.sources.forEach(s => {
      s.hoverListener = e => {
        this.drawLines(s)
      }

      s.hoverOutListener = e => {
        this.removeLines()
      }

      s.entity.addEventListener('mouseover', s.hoverListener)
      s.entity.addEventListener('mouseout', s.hoverOutListener)
    })
  }

  disconnectedCallback () {
    this.sources.forEach(s => {
      s.entity.removeEventListener('mouseover', s.hoverListener)
      s.entity.removeEventListener('mouseout', s.hoverOutListener)
    })
  }

  drawLines (source) {
    source.container.refreshDimensions()
    source.refreshDimensions()
    this.removeLines()
    const targets = toArray(this.querySelectorAll('connection-target'))
      .map(t => new ConnectionEntity(t, { container: this.container }))
      .filter(t => t.id === source.id)

    window.requestAnimationFrame(() => {
      targets.forEach(t => {
        t.refreshDimensions()
        const line = svgLine(...connectionCoordinates(source, t, { gap: this.gap }), { stroke: this.stroke, strokeWidth: this.strokeWidth })
        this.svg.appendChild(line)
        source.lines.push(line)
      })
    })
  }

  removeLines () {
    this.sources.forEach(s => {
      s.lines.forEach(l => l.remove())
    })
  }
}

customElements.define('connection-visualizer', ConnectionVisualizer)

function toArray (nodeList) {
  return Array.prototype.slice.call(nodeList)
}

function connectionCoordinates (source, target, { gap = 0 } = {}) {
  let x1, x2

  const y1 = source.offset.top + source.height / 2
  const y2 = target.offset.top + target.height / 2

  let sourceIsLeft = false
  let sourceIsRight = false

  if (source.offset.right < target.offset.left) {
    sourceIsLeft = true
    x1 = source.offset.right // use right side of source as x-axis anchor
    x2 = target.offset.left // use left side of target as x-axis anchor
  } else if (source.offset.left < target.offset.right) { // source overlaps with target on x-axis
    x1 = source.offset.left + source.width / 2 // use source middle as x-anchor
    x2 = target.offset.left + target.width / 2 // use target middle as x-anchor
  } else { // source is right of target
    sourceIsRight = true
    x1 = source.offset.left // use left side of source as x-axis anchor
    x2 = target.offset.right // use right side of target as x-axis anchor
  }

  if (sourceIsLeft) {
    x1 = x1 + gap
    x2 = x2 - gap
  } else if (sourceIsRight) {
    x1 = x1 - gap
    x2 = x2 + gap
  }

  return [x1, y1, x2, y2]
}
