/* global HTMLElement, customElements */

import Fuse from '../../../node_modules/fuse.js/dist/fuse.basic.esm.js'

import { domElement } from '../utils/dom.js'
import { debounce } from '../utils/fp.js'

const queryInputField = () => domElement(`
  <input type="text" class="fuzzy-combo-query-field form-control" />
`)

const searchResultList = () => domElement(`
  <ul class="list-group fuzzy-combo-result-list hidden card"></ul>
`)

const searchResultItem = (label, refIndex) => domElement(`
  <li class="list-group-item">
    <fuzzy-combo-result-item>
      <button data-ref-index="${refIndex}" class="btn btn-link" type="button">${label}</button>
    </fuzzy-combo-result-item>
  </li>
`)

const clearButton = (fieldLabel) => domElement(`<button class="fuzzy-combo-clear-option btn btn-link" type="button" aria-label="clear ${fieldLabel}">X</button>`)

export class FuzzyComboBox extends HTMLElement {
  constructor () {
    super()

    this.queryInputField = queryInputField()
    this.searchResultList = searchResultList()
    this.suggestionLimit = this.hasAttribute('suggestion-limit') ? parseInt(this.getAttribute('suggestion-limit')) : 10
  }

  connectedCallback () {
    this.select?.classList.add('hidden') // hide original select
    if (this.select?.hasAttribute('tabindex')) {
      this.queryInputField?.setAttribute('tabindex', this.select.getAttribute('tabindex'))
      this.select.setAttribute('tabindex', null)
    }
    this.select.after(this.queryInputField) // replace with text field for searching
    this.queryInputField.after(this.searchResultList) // append searc result list, to pick a list entry

    this.fuse = new Fuse(this.options, { keys: ['label'] })

    this.searchListener = debounce(() => this.search(), 50)
    this.queryInputField.addEventListener('input', this.searchListener)
    this.queryInputField.addEventListener('click', this.searchListener)

    const preselectedOption = this.querySelector('option[selected]')

    if (preselectedOption) {
      this.queryInputField.value = preselectedOption.label
    }

    this.selectionListener = (ev) => this.selectOption(ev.detail.refIndex)
    this.addEventListener('fuzzy-search:selected', this.selectionListener)

    this.keyboardListener = ev => {
      if (ev.keyCode === 40) {
        this.enterListByKeyboard()
      } else if (ev.keyCode === 27) {
        this.clearSearchResults()
        this.reactivateQueryFocus()
      }
    }

    this.regainFocusListener = ev => this.reactivateQueryFocus()
    this.addEventListener('fuzzy-search-item:focusout', this.regainFocusListener)

    this.cancelationListener = ev => {
      this.clearSearchResults()
      this.reactivateQueryFocus()
    }

    this.addEventListener('fuzzy-search-item:cancel', this.cancelationListener)

    this.queryInputField?.addEventListener('keydown', this.keyboardListener)

    this.focusOutListener = ev => this.focusOut(ev)
    this.addEventListener('focusout', this.focusOutListener)

    const nullOption = this.select?.querySelector('option[value=""]')
    if (nullOption) {
      this.clearOptionButton = clearButton('fubar')
      this.appendChild(this.clearOptionButton)

      this.clearButtonListener = e => {
        e.preventDefault()
        this.deselectAll()
        this.selectOption(nullOption)
      }
      this.clearOptionButton.addEventListener('click', this.clearButtonListener)
    }
  }

  disconnectedCallback () {
    this.queryInputField?.removeEventListener('input', this.searchListener)
    this.queryInputField?.removeEventListener('focus', this.searchListener)
    this.queryInputField?.removeEventListener('keydown', this.keyboardListener)
    this.removeEventListener('focusout', this.focusOutListener)
    this.removeEventListener('fuzzy-search:selected', this.selectionListener)
    this.removeEventListener('fuzzy-search-item:focusout', this.regainFocusListener)
    this.clearOptionButton?.removeEventListener('click', this.clearButtonListener)
  }

  reactivateQueryFocus () {
    if (this.queryInputField) {
      const end = this.queryInputField.value.length
      setTimeout(() => { this.queryInputField.setSelectionRange(end, end) }, 10)
      this.queryInputField.focus()
    }
  }

  focusOut (ev) {
    if (this.contains(ev.relatedTarget || ev.explicitOriginalTarget || ev.target)) {
      // click has been focussed to a sub-element, which we don't handle
      return
    }

    this.clearSearchResults()
  }

  enterListByKeyboard () {
    if (this.searchResultList.classList.contains('open')) {
      const firstItem = this.searchResultList.querySelector('fuzzy-combo-result-item')
      firstItem?.activate()
    } else {
      this.search()
    }
  }

  selectOption (itemRef) {
    const option = this.select.item(itemRef)

    this.deselectAll() // multi-selected is not supported

    if (option) {
      option.selected = true
      option.setAttribute('selected', true)
      this.queryInputField.value = option.label
    }

    this.clearSearchResults()
    this.queryInputField.focus()
  }

  get select () {
    if (this._select) {
      return this._select
    }

    this._select = this.querySelector('select')

    if (!this._select) {
      console.log('FuzzyComboBoxError: no select field found')
    }

    return this._select
  }

  get options () {
    if (this._options) {
      return this._options
    }

    this._options = Array.from(this.select?.options)

    if (!this._options) {
      console.log('FuzzyComboBoxError: no options fields found')
    }

    return this._options
  }

  search () {
    const searchTerm = this.queryInputField?.value
    this.clearSearchResults()
    if (searchTerm) {
      const results = this.fuse.search(searchTerm)

      results
        .map(r => searchResultItem(this.options[r.refIndex]?.label, r.refIndex))
        .slice(0, this.suggestionLimit) // limit to max num of list items
        .forEach(item => this.searchResultList.appendChild(item))

      this.showSearchResults()
    } else {
      const allOptions = this.options?.map((r, i) => r.label && searchResultItem(r.label, i)).filter(x => !!x)
      allOptions.forEach(item => this.searchResultList.appendChild(item))

      this.showSearchResults()
    }
  }

  clearSearchResults () {
    this.searchResultList.classList.remove('open')
    this.searchResultList.classList.add('hidden')
    this.searchResultList.innerHTML = ''
  }

  showSearchResults () {
    this.searchResultList.classList.remove('hidden')
    this.searchResultList.classList.add('open')
  }

  deselectAll () {
    this.options?.forEach(o => {
      o.removeAttribute('selected')
      o.selected = false
    })
    this.queryInputField.value = ''
  }
}

customElements.define('fuzzy-combo-box', FuzzyComboBox)
